/* uidl.c
**
** (Pseudo-)UIDL calculation
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include "mrmacs.h"
#include "tools.h"
#include "popmsg.h"
#include "folder.h"
#include "exitcodes.h"
#include "vxerror.h"
#include "inline-funcs.h"
#include "digest.h"
#include "dbif.h"

#include "cmi.h"
#include "popmsg.h"
#include "uidl.h"

#define UIDSZ (2*(sizeof (time_t) + sizeof (int)))

typedef unsigned char uchar;

static char *uidl_special (xFile fx, int msgn)
{
    DIG_CTX uidlcntxt;
    uchar *digest;
    char line[1024], *hex = "0123456789abcdef";
    char *outbuf, *outp;
    ssize_t llen, dgsz;
/*    size_t rlen = 32;*/
    int ix, lasteol, preveol;
    ifnull (uidlcntxt = digest_new ("MD5")) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "unknown digest algorithm `MD5'.\n");
    }
    dgsz = digest_size (uidlcntxt);
    digest = galloc (__FILE__, __LINE__, dgsz);
    outbuf = galloc (__FILE__, __LINE__, 2 * dgsz + 1);
    xSeek (fx, msgn);
    vxlog (5, "   calculating MD5 sum of message header ");
    lasteol = 1;
    while ((llen = xGets (line, 1024, fx)) > 0) {
	preveol = lasteol;
	digest_update (uidlcntxt, (uchar *) line, llen);
	vxlog (5, ".");
	lasteol = cuteol (line);
	if (preveol && !*line) break;
    }
    digest_final (digest, uidlcntxt);
    vxlog (5, " done.\n");
    outp = &outbuf[2*dgsz];
    *outp = '\0';
    for (ix = dgsz - 1; ix >= 0; --ix) {
	*--outp = hex[digest[ix] & 0xf];
	*--outp = hex[(digest[ix] >> 4)];
    }
    vxlog (5, "   (Pseudo-)UIDL = %s\n", outbuf);
    gfree (digest);
    digest_free (uidlcntxt);
    return outbuf;
}

/* Gib die - erwartete - Lnge des X-UIDL:-Header-Feldes zurck. */
size_t xuidl_len (int special)
{
    return 10 + (special ? 32 : UIDSZ);
}

char *uidl_calc (xFile fx, int msgn, int special)
{
    char outbuf[UIDSZ + 1];
    char *hex = "0123456789abcdef", *outp;
    int ix;
    time_t uidhigh;
    int uidlow;
    if (special) { return uidl_special (fx, msgn); }
    xSeek (fx, msgn);
    vxlog (5, "   reading UIDL from index file\n");
    flGetUID (fx, &uidhigh, &uidlow);
    outp = &outbuf[2 * sizeof (time_t)];
    for (ix = 0; ix < sizeof (time_t); ++ix) {
	*--outp = hex[uidhigh & 0xf]; uidhigh >>= 4;
	*--outp = hex[uidhigh & 0xf]; uidhigh >>= 4;
    }
    outp = &outbuf[2 * (sizeof (time_t) + sizeof (int))];
    *outp = '\0';
    for (ix = 0; ix < sizeof (int); ++ix) {
	*--outp = hex[uidlow & 0xf]; uidlow >>= 4;
	*--outp = hex[uidlow & 0xf]; uidlow >>= 4;
    }
    vxlog (5, "   UIDL = %s\n", outbuf);
    return gdupl (__FILE__, __LINE__, outbuf);
}

/* Die knstlich eingeschobene Meldung ber die Umleitung mu eine UID
** erhalten, die in jedem Fall niedriger als die aller extern empfangenen
** E-Mails ist. Da fr diese Meldung die UID nicht wichtig erscheint, kann
** ich einfach eine Konstante angeben: 0.
*/
const char *rcalc_uidl (void)
{
    static char buf[UIDSZ + 1] = { 0 };
    if (!*buf) {
	vxlog (5, "   creating (constant) UIDL for redirect message .");
	memset (buf, '0', UIDSZ);
	vxlog (5, ".");
	buf[UIDSZ] = '\0';
	vxlog (5, ". done.\n");
    }
    return buf;
}
